/*    ARGPARSE.C - Extended 'getopt' procedure plus
                   associated 'getarg' procedure

    getopt
    ------
    
    Original version:
        Keith Bostic
        
    Extended version:
        Lloyd Zusman
        Masterbyte Software
        Los Gatos, CA
        
    Adapted to OS9/68K standard:
        Martin Gregorie    
        Harlow,    
        U.K.
    
    'getopt' must be called first to parse options on the command
    line. The extensions to OS/9 standards add the facilities:
    1   Argument options must meet one of the following standards: 
            -xval  
            -x=val  
            -x= val
    2   The entire option or just the value may be enclosed in quotes
    3   Options not taking values can be concatenated
    4   Added the optcase global; 1=case sensitive 0=ignore case.
        Defaults to case sensitive.
    5   Returns:
            NULL for non-options (skips other arguments)
            'optchar' for valid options
            EOF for end of command line (all options processed)
    6   Causes 'getarg' to ignore parameters that have already
        been processed as options and/or option values
                                      
    getarg
    ------
    
    Original version:
        Martin Gregorie    
        Harlow,                
        U.K.
    
    'getarg' returns command line parameters ignored by 'getopt'.
    It must always be called after 'getopt' has completely scanned
    the command parameters. optind should be reset to 1 before the first
    call. It returns:
        0    if the argument has been processed as an option or
             option value by getopt()
        1    argument found. optarg points to the argument.
        EOF  all arguments processed.
        
    
    NOTE
    ----
    This version can ONLY be compiled under Linux, as the _tell()
    function uses 'writeln' to force a newline onto the terminal; 
    'writeln' is not a standard UNIX function.
*/

#include <stdio.h>
#include <io.h>
#include <string.h>
#include "environ.h"

/*
    Constants that are not referenced in the executable portion
    of the code ... their sole purpose is to initialise global
    variables.
*/

#define BADCH        (int)'?'
#define ERRFD        2
#define MAYBESEP     (int)'\0'
#define NEEDSEP      (int)':'
#define START        "-"

/*    Other constant definitions */

#define EMSG        ""
#define PARMSTART   (int)'='
#define FOUND       1
#define NOTHING     0
#define MAXUSED     25
#define HUGEVAL     9999
#define QUOTES      "'\""

#ifndef EOF
#    define EOF (-1)
#endif    /* ! EOF */

#ifndef NULL
#    define NULL    (char *)0
#endif    /* ! NULL */

/*    
    Here are all the global variables that control getopt().
    They are accessable by the caller who may change the 
    standard defaults shown below.
*/

char *optstart = START;    /* list of characters that start options */
int  optbad = BADCH;       /* character returned on error */
int  opterr = 1;           /* if true, output error message */
int  opterrfd = ERRFD;     /* file descriptor for error text */
int  optind = 1;           /* index into parent argv vector */
int  optmaybe = MAYBESEP;  /* flag for optional argument */
int  optneed = NEEDSEP;    /* flag for mandatory argument */
int  optcase = 1;          /* 1=case sensitive options; 0=caseless */

/*
    The following global contains NULL or a pointer to the 
    option value (after getopt()) or the argument that has been 
    found (getarg()).
*/
char *optarg;        /* argument associated with option */

/*
    Internal globals with externally visible names. 
    This follows the standard Unix getopt()
*/
int  optchar = 0;    /* flag character for the current option argument */
int  optopt;         /* option character checked by last getopt() call */

/*    
    Private global variables used by 'getopt' and 'getarg'
    for intercommunication.
*/ 

int  _used = 0;            /* pointer into _usedarg array */
int  _usedarg[MAXUSED];    /* list of indices of command params 
                              used as arguments of options */

/*
    Private global used to keep the option scan pointer tidy.

    The option scanner points here when there's no 
    current option argument.
*/
char *_emsgstr = EMSG;

/*    
    General support procedures
    --------------------------
    
    _tell
    -----
    Conditionally output an error message.
    
*/
void _tell(char *s)
{
    char    option;
    char*   name;
    
    if (opterr && opterrfd >= 0)
    {
        option = (char)optopt;
        name = (char*)progname();
        write(opterrfd, name, strlen(name));
        write(opterrfd, (s), strlen(s));
        write(opterrfd, &option, 1);
        write(opterrfd, "\n", 1);
    }
}

/*    
    _sindex 
    -------
    Used to determine whether a given character is in 
    a list. It returns a pointer to the character in 
    'string' if it matches 'ch' and NULL otherwise.
*/
char *_sindex(const char *string, int ch)
{
    if (string)
        return (index(string, ch));
    return(NULL);
}

/*
    Extended getopt function
    ------------------------
*/
int getopt(int nargc, char * const nargv[], const char *ostr)
{
    static char    *place = NULL;    /* option letter processing */
    register char  *oli;             /* option letter list index */
    register char  *osi = NULL;      /* option start list index */
    int            expect = 0;       /* set when '=' found */
    char           *qti;             /* vars for dequoting args */
    int            qlen;
    char           *quotes = QUOTES;
    
    /* Various checks for the end of the command line */

    if (!nargv)
        return(EOF);
    if (optind >= nargc)    /* all args processed */
        return(EOF);
    if (!nargv[optind])     /* null arg found */
        return(EOF);
        
    if (!place)              /* place always points to a string */
        place = _emsgstr;    /* even if it is null */ 
        
    /* 
        initialise the used arguments list if the scan
        is starting from the beginning of the command
        parameters
    */
    
    if (optind <= 1)
    {
        for ( _used = 0; _used < MAXUSED; _used++) 
            _usedarg[_used] = HUGEVAL;
        _used = 0;
    }
        
    /*    Update scanning pointer; we need to start a new argument */

    if (!*place)
    {
        place = nargv[optind];
        if (!place 
            || optind >= nargc 
            || !*place)
            return(EOF);
        if (osi = _sindex(optstart, *place))
            optchar = (int)*osi;
        if (osi == NULL)        /* not an option argument */
        {
            place = _emsgstr;
            optind++;
            return(0);
        }
        place++;
            
        /* 
            Two adjacent, identical flag characters were found
            at the start of an option argument. They are
            treated as an unconditional end of option scan.
            This may make trouble for subsequent use of getarg()
            if it is followed by non-options starting with a
            flag character, which getarg() will skip. 
            
            You have been warned!
            
            This action is disabled by putting the start
            character in the valid options list and processing
            it as an option in its own right.
        */
        if (*place == place[-1])
        {
            ++optind;
            place = _emsgstr;
            return(EOF);
        }
    }
    
    /*
        If the option is a separator or the option isn't in the list,
        we've got an error.
    */
    optopt = (int)*place++;
    if (!optcase)
        if (optopt>='A' && optopt<='Z')
            optopt += 'a'-'A';
    oli = _sindex(ostr, optopt);
    if (optopt == optneed || optopt == optmaybe || !oli)
    {
        /*
            If we're at the end of the current argument,
            increment the argument index.
        */
        if (!*place)
            ++optind;
        _tell(": illegal option - ");    /* bye bye */
        return (optbad);
    }

    /*
        If there is no argument indicator, then we don't even
        try to return an argument.
    */
    if (!*++oli || (*oli != optneed && *oli != optmaybe))
    {
        /*
            If we're at the end of the current argument,
            increment the argument index.
        */
        if (!*place)
            ++optind;
        optarg = NULL;
    }
    else
    {
        /*
            If we're here theres an argument indicator. 
            It's handled differently depending on whether 
            it's a mandatory or an optional argument
    
            There are several possibilities:
            1   the argument directly follows the option
            2   this is the last char in the param string,
                which indicates:
                a    either a missing mandatory arg (error!)
                b    an absent optional arg (ok)
            3   '=' directly follows the option in which 
                case the arg may:
                a    directly follow the '=' (ok)
                b    form the next param (ok) This case is 
                     allowed for MS-DOS compatibility
                c    be missing (error!)
        */
        if (*place == PARMSTART)
        {
            expect = *place;    /* case 3 recognised and */
            place++;            /* noted as such for later */
        }

        if (*place)
            optarg = place;     /* cases 1 and 3a */
        else
        {                       /* cases 2a, 2b, 3b, 3c */
            if (!expect)
            {
                if (*oli == optneed)
                {
                    place = _emsgstr;    /* case 2a */
                    optind++;
                    _tell(": option requires an argument - ");
                    return (optbad);
                }
                else
                    optarg = NULL;    /* case 2b */
            }
            else
            {
                optind++;
                if (nargc <= optind || !nargv[optind])
                {
                    _tell(": option requires an argument - "); /* case 3c */
                    return (optbad);
                }
                else
                {
                    /*    case 3b
                    
                        separated argument. Record it in _usedarg to
                        so that 'getarg' knows it is part of an 
                        option
                    */
                    if (_used < MAXUSED)
                        _usedarg[_used++] = optind;
                    else
                    {
                        _tell(": too many separated arguments -");
                        return (optbad);
                    }
                        
                    optarg = nargv[optind];    /* case 3b */
                }
            }
        }
        
        place = _emsgstr;
        ++optind;
    }
    
    /* remove surrounding quotes (if any) from the argument */
    
    if (optarg && (qti = _sindex(quotes, *optarg)))
    {
        qlen = strlen(optarg);
        optarg++;            /* skip leading quote */
        optarg[qlen-2] = 0;    /* and null out trailing one */
    }

    return(optopt);        /*    return option letter */
}

/*
    getarg function
    ---------------
*/
int getarg(int nargc, char **nargv)
{
    register char     *osi = NULL;    /* option start list index */
    int               reply;
    static int        scanused = 0;   /* readout index for _usedarg */
    
    if (!nargv)                       /* check for end of param list */
        return(EOF);
    if (nargc <= optind || !nargv[optind])
        return(EOF);
    if (optind >= nargc)
        return(EOF);
        
    if (optind <= 1)                  /* reset _usedarg scan if needed */
        scanused = 0;
        
     /*    
         Find the first relevant reference in the array. Its initialised 
         to high values, so we want the first entry thats not less than 
         optind. If theres a match then the argument was marked as part 
         of an option by getopt() and will be ignored.
     */
    while (_usedarg[scanused] < optind)
        scanused++;
        
    if (_usedarg[scanused] == optind)
    {                                 /* argument already used as a */
        optarg = _emsgstr;            /* separated argument to an   */
        reply = NOTHING;              /* an option                  */
    }
    else
    {        
        /* 
            See if the command line parameter is an option 
        */
        osi = _sindex(optstart, *nargv[optind]); 

        if (osi == NULL)
        {                            /* not an option */
            optarg = nargv[optind];
            reply = FOUND;
        }
        else
        {                            /* argument is an option */
            optarg = _emsgstr;
            reply = NOTHING;
        }
    }

    optind++;
    return(reply);
}
